// Extract.cpp

#include "StdAfx.h"

#include "../../UI/Common/Extract.h"					//	pXύX

#include "Common/StdOutStream.h"
#include "Windows/Defs.h"
#include "Windows/PropVariant.h"
#include "Windows/FileDir.h"

#include "../../UI/Common/OpenArchive.h"				//	pXύX

#include "../../UI/Console/OpenCallbackConsole.h"		// ǉ
#include "../../UI/Console/ExtractCallbackConsole.h"	// ǉ

#ifndef EXCLUDE_COM
#include "Windows/DLL.h"
#endif

using namespace NWindows;

HRESULT DecompressArchive(
    IInArchive *archive,
    const UString &defaultName,
    const NWildcard::CCensorNode &wildcardCensor,
    const CExtractOptions &options,
    IExtractCallbackUI *callback)
{
  CRecordVector<UInt32> realIndices;
  UInt32 numItems;
  RINOK(archive->GetNumberOfItems(&numItems));

  for(UInt32 i = 0; i < numItems; i++)
  {
    UString filePath;
    RINOK(GetArchiveItemPath(archive, i, options.DefaultItemName, filePath));
    bool isFolder;
    RINOK(IsArchiveItemFolder(archive, i, isFolder));
    if (!wildcardCensor.CheckPath(filePath, !isFolder))
      continue;
    realIndices.Add(i);
  }
  if (realIndices.Size() == 0)
  {
    callback->ThereAreNoFiles();
    return S_OK;
  }

  CArchiveExtractCallback *extractCallbackSpec = new CArchiveExtractCallback;
  CMyComPtr<IArchiveExtractCallback> extractCallback(extractCallbackSpec);
  
  UStringVector removePathParts;

  UString outDir = options.OutputDir;
  outDir.Replace(L"*", defaultName);
  if(!outDir.IsEmpty())
    if(!NFile::NDirectory::CreateComplexDirectory(outDir))
    {
      throw UString(L"Can not create output directory ") + outDir;
    }

  g_StdOut.GetProgressDialog()->SetArchiveInfo(archive, outDir, options.PathMode, removePathParts);		// ǉ removePathParts͕KvC邪ꉞ

  extractCallbackSpec->Init(
      archive, 
      callback,
      options.StdOutMode,
      outDir, 
      options.PathMode, 
      options.OverwriteMode,
      removePathParts, 
      options.DefaultItemName, 
      options.ArchiveFileInfo.LastWriteTime,
      options.ArchiveFileInfo.Attributes);

  /* 폜
  HRESULT result = archive->Extract(&realIndices.Front(), 
    realIndices.Size(), options.TestMode? 1: 0, 
      extractCallback);
     폜܂ */

  /* ǉ */
  HRESULT result;
  int nRetry = 0;
  bool PasswordIsDefined = ((CExtractCallbackConsole*)callback)->PasswordIsDefined;
  while((result = archive->Extract(&realIndices.Front(), 
    realIndices.Size(), options.TestMode? 1: 0, 
      extractCallback)) == ERROR_PASSWORD_FILE)
  {
	if (PasswordIsDefined)
	  break;
    CMyComBSTR password;
    RINOK(extractCallbackSpec->CryptoGetTextPassword(&password));
    ++nRetry;
    if (*(const wchar_t *)password == '\0' || nRetry == 3)
      break;
    CPasswordDialog dlg(FALSE);
    if (dlg.DoModal() == IDCANCEL)
    {
      result = ERROR_PASSWORD_FILE;
      break;
    }
    callback->SetPassword(dlg.GetPassword());
  }
  /* ǉ܂ */

  return callback->ExtractResult(result);
}

HRESULT DecompressArchives(
    UStringVector &archivePaths, UStringVector &archivePathsFull,    
    const NWildcard::CCensorNode &wildcardCensor,
    const CExtractOptions &optionsSpec,
    IOpenCallbackUI *openCallback,
    IExtractCallbackUI *extractCallback)
{
  CExtractOptions options = optionsSpec;
  for (int i = 0; i < archivePaths.Size(); i++)
  {
    const UString &archivePath = archivePaths[i];
    NFile::NFind::CFileInfoW archiveFileInfo;
    if (!NFile::NFind::FindFile(archivePath, archiveFileInfo))
      throw "there is no such archive";

    if (archiveFileInfo.IsDirectory())
      throw "there is no such archive";

    options.ArchiveFileInfo = archiveFileInfo;

    RINOK(extractCallback->BeforeOpen(archivePath));
    CArchiveLink archiveLink;
	bool PasswordIsDefined = ((COpenCallbackConsole*)openCallback)->PasswordIsDefined;		// ǉ
    HRESULT result = MyOpenArchive(archivePath, archiveLink, openCallback);
	/* ǉ */
	for (int nRetry = 1; result != S_OK && nRetry < 3; ++nRetry)
	{
		UString password;
		openCallback->GetPasswordIfAny(password);
		if (PasswordIsDefined || password.IsEmpty())
			break;
        ((COpenCallbackConsole*)openCallback)->PasswordIsDefined = false;
		result = MyOpenArchive(archivePath, archiveLink, openCallback);
	}
	/* ǉ܂ */
    RINOK(extractCallback->OpenResult(archivePath, result));
    if (result != S_OK)
      continue;

    for (int v = 0; v < archiveLink.VolumePaths.Size(); v++)
    {
      int index = archivePathsFull.FindInSorted(archiveLink.VolumePaths[v]);
      if (index >= 0 && index > i)
      {
        archivePaths.Delete(index);
        archivePathsFull.Delete(index);
      }
    }
    UString password;
    RINOK(openCallback->GetPasswordIfAny(password));
    if (!password.IsEmpty())
    {
      RINOK(extractCallback->SetPassword(password));
    }

    options.DefaultItemName = archiveLink.GetDefaultItemName();
    RINOK(DecompressArchive(
        archiveLink.GetArchive(), archiveLink.GetDefaultItemName(),
        wildcardCensor, options, extractCallback));
  }
  return S_OK;
}